#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "Constants.h"
#include "Cpu.h"
#include "Memory.h"
#include "Io.h"
#include "Log.h"
#include "DataBus.h"
#include "Sound.h"
#include "Statistics.h"

#include "CpuUtilities.h"
#include "CpuNoPrefix.h"
#include "CpuEDPrefix.h"
#include "CpuCBPrefix.h"
#include "CpuDDAndFDPrefix.h"
#include "CpuDDCBAndFDCBPrefix.h"


static struct regs* _regs = NULL;
static Uint16 _savedPc;
static Uint8 _runMode;
static Uint8 _isHalted;
static Uint8 _cpuStarted = 0;
static Uint8 _logEveryInstruction = 0;
static Uint8 _useDataBusIM2HandlerLSB = 0;

static Uint32 _totalTStates = 0;	// total elapsed tstates

static volatile Uint8 _interruptRequested;
static volatile enum interruptMode _interruptMode;

static Uint32 _executedTStates;
static Uint32 _executionTStateTarget;

#define _DEBUG_BUFFER_SIZE 256
static const char* _debugBuffer[_DEBUG_BUFFER_SIZE];


struct regNames _regNames = { .A = "A", .F = "F", .B = "B", .C = "C", .D = "D", .E = "E", .H = "H", .L = "L", .AF = "AF", .BC = "BC", .DE = "DE", .HL = "HL", .IXL = "IXL", .IXH = "IXH",
	.IX = "IX", .IYL = "IYL", .IYH = "IYH", .IY = "IY", .R = "R", .I = "I", .IR = "IR", .SP = "SP", .PC = "PC", .IFF = "IFF", .A_alt = "A'", .B_alt = "B'", .F_alt = "F'", .C_alt = "C'", .D_alt = "D'",
	.E_alt = "E'", .H_alt = "H'", .L_alt = "L'" };

struct regNames _regNamesReference = { .A = "(A)", .F = "(F)", .B = "(B)", .C = "(C)", .D = "(D)", .E = "(E)", .H = "(H)", .L = "(L)", .AF = "(AF)", .BC = "(BC)", .DE = "(DE)", .HL = "(HL)", .IXL = "(IXL)", .IXH = "(IXH)",
	.IX = "(IX)", .IYL = "(IYL)", .IYH = "(IYH)", .IY = "(IY)", .R = "(R)", .I = "(I)", .IR = "(IR)", .SP = "(SP)", .PC = "(PC)", .IFF = "(IFF)", .A_alt = "(A')", .B_alt = "(B')", .F_alt = "(F')", .C_alt = "(C')", .D_alt = "(D')",
	.E_alt = "(E')", .H_alt = "(H')", .L_alt = "(L')" };

struct regs* _cpu_make_regs() {
	struct regs* value = (struct regs*)malloc(1 * sizeof(struct regs));
	if (value == NULL) {
		return NULL;
	}

	// x86 is little endian, so LSB must come first
	value->AF = (Uint16*)&value->F;
	value->BC = (Uint16*)&value->C;
	value->DE = (Uint16*)&value->E;
	value->HL = (Uint16*)&value->L;
	value->IR = (Uint16*)&value->R;
	value->IX = (Uint16*)&value->IXL;
	value->IY = (Uint16*)&value->IYL;

	value->AF_alt = (Uint16*)&value->F_alt;
	value->BC_alt = (Uint16*)&value->C_alt;
	value->DE_alt = (Uint16*)&value->E_alt;
	value->HL_alt = (Uint16*)&value->L_alt;

	*value->AF = 0xFFFF;
	value->PC = 0;
	value->SP = 0xFFFF;
	value->IFF = 0;
	value->I = 0;
	value->R = 0;

	*value->AF_alt = 0xFEDC;
	*value->BC_alt = 0xBA98;
	*value->DE_alt = 0x7654;
	*value->HL_alt = 0x3210;

	return value;
}

const char* cpu_prefix_to_string(enum prefixType prefix) {
	switch (prefix) {
	case NoPrefix: return "None";
	case DDCB: return "DDCB";
	case DD: return "DD";
	case FDCB: return "FDCB";
	case FD: return "FD";
	case ED: return "ED";
	case CB: return "CB";
	default: return "Unknown";
	}
}

void cpu_start(Uint8 logEveryInstruction, Uint8 useDataBusIM2HandlerLSB) {
	cpu_destroy();
	_regs = _cpu_make_regs();
	_runMode |= CPU_RUN_MODE_FLAG__EXECUTE;
	_isHalted = 0;
	_logEveryInstruction = logEveryInstruction;
	_useDataBusIM2HandlerLSB = useDataBusIM2HandlerLSB;

	_interruptMode = Mode0;
	_interruptRequested = 0;
	_totalTStates = 0;
	_cpuStarted = 1;
}

void cpu_destroy() {
	if (_regs != NULL) {
		free(_regs);
		_regs = NULL;
	}
	_cpuStarted = 0;
}

void cpu_request_interrupt() {
	_interruptRequested = 1;
}

void cpu_set_interrupt_mode(enum interruptMode mode) {
	_interruptMode = mode;
}

struct regs* cpu_regs() {
	return _regs;
}

struct instructionOpcode cpu_resolve_instruction_opcode() {
	enum prefixType prefix = NoPrefix;
	Uint16 tstates = 0;
	Uint16 startPc = _regs->PC;

	while (1) {
		if (prefix == NoPrefix) {
			Uint8 first = memory_read8(_regs->PC++);
			tstates += 4;	// M1 cycle

			// see if there's a prefix
			if (first == 0xDD) {
				Uint8 second = memory_read8(_regs->PC++);

				if (second == 0xCB) {
					tstates += 4;	// M1 cycle
					// DDCB prefix
					// 
					// CB and opcode will form instructions similar to those in the CB - prefixed opcodes section above.
					// However, these will now operate on(IX + d)and, if the instruction isn't BIT, copy the result to the register they would have initially acted upon, unless it was (HL).
					prefix = DDCB;
					continue;
				}
				else if (second == 0xDD || second == 0xED || second == 0xFD) {
					// a new prefix follows, so this one is a NOOP (though it still takes some tstates)

					// NOTE: we leave PC on this second opcode
					_regs->PC--;
					prefix = NoPrefix;
					continue;
				}
				else {
					// DD prefix
					// 
					// If the next opcode makes use of HL, H, L, but not (HL), any occurrence of these will be replaced by IX, IXH, IXL respectively.An exception of this is EX DE, HL which is unaffected.
					// If the next opcode makes use of(HL), it will be replaced by(IX + d), where d is a signed 8 - bit displacement immediately following the opcode(any immediate data, i.e.n, will follow the displacement byte), and any other instances of Hand L will be unaffected.Therefore, an instruction like LD IXH, (IX + d) does not exist, but LD H, (IX + d) does.
					// All other instructions are unaffected.
					prefix = DD;
					// NOTE: we leave PC on this second opcode
					_regs->PC--;
					continue;
				}
			}
			else if (first == 0xFD) {
				Uint8 second = memory_read8(_regs->PC++);

				if (second == 0xCB) {
					// FDCB prefix
					//
					// like DDCB prefix, except they use IY and not IX
					tstates += 4;	// M1 cycle
					prefix = FDCB;
					continue;
				}
				else {
					// FD prefix
					// The FD prefix acts exactly like the DD prefix, but the IY register is used instead of IX.
					// NOTE: we leave PC on this second opcode
					_regs->PC--;
					prefix = FD;
					continue;
				}
			}
			else if (first == 0xED) {
				// ED prefix
				prefix = ED;
				continue;
			}
			else if (first == 0xCB) {
				// CB prefix
				prefix = CB;
				continue;
			}

			// this is not a prefix
			_regs->PC--;		// first byte was not the start of a prefix, so don't count it yet
			tstates -= 4;
			// flow into below, allowing the instruction opcode to be non-prefixed
		}

		// at this point we may or may not have a prefix
		// but we KNOW that the byte at PC is either:
		//     - a mandatory displacement byte, when prefix is DDCB or FDCB, or
		//     - an instruction opcode
		Uint16 displacementPc = _regs->PC;
		Uint8 hasDisplacementByte = 0;
		Uint8 displacementByte = 0;
		if (prefix == DDCB || prefix == FDCB) {
			// PC is on a displacement byte
			hasDisplacementByte = 1;
			displacementByte = memory_read8(_regs->PC++);
			tstates += 4;	// M1 cycle
			// PC is now on the opcode byte
		}

		// here, the byte at PC is an instruction opcode
		Uint8 first = memory_read8(_regs->PC);
		tstates += 4;	// M1 cycle

		struct instructionOpcode result;
		result.opcode = first;
		result.tstatesElapsed = tstates;
		result.opcodePc = _regs->PC;
		result.prefix = prefix;
		result.prefixString = cpu_prefix_to_string(result.prefix);
		result.startPc = startPc;

		result.displacementPc = displacementPc;
		result.displacement = displacementByte;
		result.hasDisplacement = hasDisplacementByte;

		_regs->PC++;
		return result;
	}
}

// creates a partially-populated instruction from only what is known about
// what has been decoded up to and including the instruction opcode
struct instruction* _cpu_inst_opcode__to__instruction__preliminary(struct instructionOpcode* instructionOpcode) {
	struct instruction* instruction = (struct instruction*) malloc(1 * sizeof(struct instruction));
	if (instruction == NULL) {
		return NULL;
	}
	instruction->startPc = instructionOpcode->startPc;

	instruction->prefix.type = instructionOpcode->prefix;
	instruction->prefix.humanReadable = instructionOpcode->prefixString;

	instruction->opcode.pc = instructionOpcode->opcodePc;
	instruction->opcode.value = instructionOpcode->opcode;

	// while there might be a displacement BEFORE the instruction opcode (in which case we'd have it
	// here), most instructions will change this based on whether they have a displacement
	// AFTER the opcode
	instruction->displacement.hasDisplacement = instructionOpcode->hasDisplacement;
	instruction->displacement.pc = instructionOpcode->displacementPc;
	instruction->displacement.value = instructionOpcode->displacement;

	// so far it's taken as much as reading up to and including the opcode has taken
	instruction->tstatesElapsed = instructionOpcode->tstatesElapsed;

	instruction->dissassembledName = NULL;
	cpu_set_instruction_disassembled_name(instruction, "unspecified");

	// assume decode for now - if execution is needed, microcode will set it as such
	instruction->outcome = DecodedOnly;
	// we don't know this yet - filled in once the instruction is fully decoded
	instruction->immediate.type = NoImmediate;

	return instruction;
}

// entry point for running the next CPU instruction
struct instruction* cpu_run_next_instruction() {
	if (_isHalted) {
		struct instruction* halted = (struct instruction*)malloc(1 * sizeof(struct instruction));
		if (halted == NULL) {
			return NULL;
		}
		halted->outcome = SkippedDueToHalt;
		halted->dissassembledName = (char*)malloc(1 * sizeof(char*));
		if (halted->dissassembledName == NULL) {
			return NULL;
		}
		*halted->dissassembledName = 0;
		return halted;
	}

	struct instructionOpcode instructionOpcode = cpu_resolve_instruction_opcode();
	data_bus_write8(instructionOpcode.opcode);

	struct instruction* instruction = _cpu_inst_opcode__to__instruction__preliminary(&instructionOpcode);
	switch (instruction->prefix.type) {
	case NoPrefix:
		cpu_run_no_prefix(instruction);
		break;
	case ED:
		cpu_run_ED_prefix(instruction);
		break;
	case CB:
		cpu_run_CB_prefix(instruction);
		break;
	case DD:
		cpu_run_DD_and_FD_prefix(instruction);
		break;
	case FD:
		cpu_run_DD_and_FD_prefix(instruction);
		break;
	case DDCB:
		cpu_run_DDCB_and_FDCB_prefix(instruction);
		break;
	case FDCB:
		cpu_run_DDCB_and_FDCB_prefix(instruction);
		break;
	default:
		instruction->outcome = UnknownInstruction;
		break;
	}

	if (instruction->outcome == Executed) {
		sound_add_sample(io_is_speaker_active(), _totalTStates);
		statistics_record(instruction);
	}

	if (instruction->outcome == UnknownInstruction) {
		sprintf_s((char* const)_debugBuffer, _DEBUG_BUFFER_SIZE - 10, "%04x: Unrecognized instruction with prefix: %s   opcode: %02x", instruction->startPc, instruction->prefix.humanReadable, instruction->opcode.value);
		log_write((char*)_debugBuffer);

		char* message = (char*)cpu_make_regs_dump_string__static_mem_ptr();
		log_write(message);
	}

	if (_logEveryInstruction) {
		sprintf_s((char* const)_debugBuffer, _DEBUG_BUFFER_SIZE - 10, "%04x\t%s\n", instruction->startPc, instruction->dissassembledName);
		log_write((char*)_debugBuffer);

		char* message = (char*)cpu_make_regs_dump_string__static_mem_ptr();
		log_write(message);
	}

	return instruction;
}

void cpu_begin_decode_only_run() {
	_savedPc = _regs->PC;
	_runMode &= ~CPU_RUN_MODE_FLAG__EXECUTE;
}

void cpu_end_decode_only_run() {
	_regs->PC = _savedPc;
	_runMode |= CPU_RUN_MODE_FLAG__EXECUTE;
}

void cpu_destroy_instruction(struct instruction* instruction) {
	if (instruction == NULL) {
		return;
	}
	if (instruction->dissassembledName != NULL) {
		free(instruction->dissassembledName);
		instruction->dissassembledName = NULL;
	}
	free(instruction);
}

void _cpu_handle_interrupt() {
	Uint16 handlerPtrPtr, handlerPtr;

	if (!_interruptRequested) {
		return;
	}
	if (!cpu_regs()->IFF) {
		// ignore interrupt requests while interrupts are disabled
		_interruptRequested = 0;
		return;
	}

	// an interrupt has occurred
	switch (_interruptMode) {
	case Mode0:
		// unsupported... ZX Spectrum data bus is unreliable when the CPU handles an interrupt
		break;
	case Mode1:
		cpu_push_PC();
		cpu_regs()->PC = 0x38;
		break;
	case Mode2:
		handlerPtrPtr = ((Uint16)cpu_regs()->I) << 8;
		if (_useDataBusIM2HandlerLSB) {
			handlerPtrPtr |= data_bus_read8();
		} else {
			handlerPtrPtr |= 0xff;
		}
		handlerPtr = memory_read16(handlerPtrPtr);

		cpu_push_PC();
		cpu_regs()->PC = handlerPtr;
		break;
	}

	_interruptRequested = 0;
	_isHalted = 0;
}

// this is used by string instructions to know when they can exit
Uint8 cpu_can_continue_string_operation(Uint32 currentTStatesConsumedByInstruction) {
	if (_executedTStates + currentTStatesConsumedByInstruction < _executionTStateTarget) {
		return 1;
	}

	// string operation has consumed all of our remaining tstate target,
	// so it cannot continue
	return 0;
}

Uint32 cpu_run_many(Uint32 tstateTarget) {
	_executedTStates = 0;
	_executionTStateTarget = tstateTarget;

	_cpu_handle_interrupt();

	if (_isHalted) {
		return _executedTStates;
	}

	// run instructions until we pass the target
	while (_executedTStates < _executionTStateTarget) {
		if (_isHalted) {
			return _executedTStates;
		}

		struct instruction* instruction = cpu_run_next_instruction();
		if (instruction->outcome == Executed) {
			// instruction has executed successfully

			_executedTStates += instruction->tstatesElapsed;
			_totalTStates += instruction->tstatesElapsed;
		}

		cpu_destroy_instruction(instruction);
	}

	return _executedTStates;
}

Uint8 cpu_mode() {
	return _runMode;
}

void cpu_halt() {
	_isHalted = 1;
}
