#include <Windows.h>
#include <SDL.h>
#include <SDL_timer.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys\types.h> 
#include <sys\stat.h>

#include "Constants.h"
#include "Video.h"
#include "Memory.h"
#include "Cpu.h"
#include "Tests.h"
#include "Io.h"
#include "Keyboard.h"
#include "Kempston.h"
#include "Log.h"
#include "Sound.h"
#include "Statistics.h"

#include "SnaLoader.h"

static Uint8 _zoom = DEFAULT_ZOOM;
static Uint8 _soundOn = 1;

void rom_load() {
    log_write("Starting ROM load");

    struct stat info;
    const char* filename = "48k.bin";
    log_write((char*)filename);
    if (stat(filename, &info) != 0) {
        log_write("Error: ROM file 48k.bin not found");
        return;
    }

    char* file = (Uint8*)malloc(info.st_size);
    if (file == NULL) {
        log_write("Error: unable to allocate memory for ROM file");
        return;
    }
    FILE* fp;
    errno_t result = fopen_s(&fp, filename, "rb");
    if (result) {
        log_write("Error: could not open ROM file");
        return;
    }
    // try to read a single block of info.st_size bytes
    size_t blocks_read = fread(file, info.st_size, 1, fp);
    if (blocks_read != 1) {
        log_write("Error: could not read from SNA file");
        return;
    }
    fclose(fp);

    memory_load(file, 0, SPECTRUM_ROM_SIZE);
    free(file);
    log_write("Loaded ROM into ZX Spectrum memory");
}

int main(int argc, char* argv[]) {
    char* snaFile = NULL;
    Uint8 snaFound = 0;

    SDL_Log("");
    SDL_Log("zxian v4 - a ZX Spectrum emulator written by Sebastian Mihai");

#ifdef _DEBUG
    HWND windowHandle = GetConsoleWindow();
    ShowWindow(windowHandle, SW_SHOW);
#endif // _DEBUG

    SDL_Log("Usage:");
    SDL_Log("         zxian -switch value");
    SDL_Log("Switches:");
    SDL_Log("  -sna <SNA file>     loads the specified SNA file");
    SDL_Log("                      example:    -sna manicminer.sna");
    SDL_Log("  -zoom <integer>     enlarges window by an integer factor");
    SDL_Log("                      example:    -zoom 2");
    SDL_Log("  -silence            silences sounds");
    SDL_Log("");
    SDL_Log("Advanced switches:");
    SDL_Log("  -instructionlog     logs every instruction executed by");
    SDL_Log("                      the CPU; produces VERY large log files");
    SDL_Log("  -instructionstats   logs run-time instruction statistics");
    SDL_Log("                      such as instruction frequency, etc.");
    SDL_Log("  -im2lsb             read LSB from data bus when finding");
    SDL_Log("                      IM2 handler address, instead of ");
    SDL_Log("                      hardcoding 0xFF; some games may fail");

    // find SNA
    for (int i = 0; i < argc-1; i++) {
        if (!strcmp(argv[i], "-sna")) {
            snaFile = argv[i + 1];
            snaFound = 1;
        }
    }

    // find zoom
    for (int i = 0; i < argc - 1; i++) {
        if (!strcmp(argv[i], "-zoom")) {
            _zoom = atoi(argv[i + 1]);
            if (_zoom < 1) {
                _zoom = DEFAULT_ZOOM;
            }
        }
    }

    // find silence
    for (int i = 0; i < argc; i++) {
        if (!strcmp(argv[i], "-silence")) {
            _soundOn = 0;
        }
    }

    Uint8 instructionLog = 0;
    // find instructionlog
    for (int i = 0; i < argc; i++) {
        if (!strcmp(argv[i], "-instructionlog")) {
            instructionLog = 1;
        }
    }

    Uint8 instructionStats = 0;
    // find instructionlog
    for (int i = 0; i < argc; i++) {
        if (!strcmp(argv[i], "-instructionstats")) {
            instructionStats = 1;
        }
    }

    Uint8 useDataBusIM2HandlerLSB = 0;
    // find useDataBusIM2HandlerLSB
    for (int i = 0; i < argc; i++) {
        if (!strcmp(argv[i], "-im2lsb")) {
            useDataBusIM2HandlerLSB = 1;
        }
    }

    log_start();

    SDL_Event event;
    SDL_Window* window;
    log_write("Created SDL window");
    
    statistics_start();
    keyboard_start();
    kempston_start();
    io_start();
    memory_start();
    rom_load();
    
    // these are used to test video rendering by loading a screen and then entering
    // an infinite loop
    //tests_screen_load();
    //tests_rom_load();
    
    // the unit tests exercise parsing of instruction opcodes and the execution
    // tests are used to test one or few instructions at a time
    //tests_run_unit_tests();
    //tests_run_execution_tests();

    cpu_start(instructionLog, useDataBusIM2HandlerLSB);
    if (snaFound) {
        sna_load(snaFile);
    }

    SDL_Init(SDL_INIT_VIDEO);

    int windowWidth = VISIBLE_SPECTRUM_SCREEN_WIDTH * _zoom;
    int windowHeight = VISIBLE_SPECTRUM_SCREEN_HEIGHT * _zoom;
    window = SDL_CreateWindow("zxian v4 (by Sebastian Mihai)", SDL_WINDOWPOS_CENTERED, SDL_WINDOWPOS_CENTERED, windowWidth, windowHeight, 0);
    
    if (_soundOn) {
        sound_start();
    }
    video_start(window, _zoom);

    while (1) {
        SDL_Delay(10);
        SDL_PollEvent(&event);
        
        if (event.type == SDL_QUIT) {
            break;
        }

        if (event.type == SDL_KEYDOWN) {
            keyboard_keydown(event.key.keysym);
            kempston_keydown(event.key.keysym);
        }

        if (event.type == SDL_KEYUP) {
            keyboard_keyup(event.key.keysym);
            kempston_keyup(event.key.keysym);
        }
    }

    log_write("Exit was requested");

    video_destroy();
    sound_destroy();

    SDL_DestroyWindow(window);
    SDL_Quit();

    cpu_destroy();
    memory_destroy();
    io_destroy();
    kempston_destroy();
    keyboard_destroy();

    if (instructionStats) {
        statistics_log_all();
    }
    statistics_destroy();

    log_destroy();

    return 0;
}
